/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
/*
 * Uses the mountlist library in gnulib to map system device IDs and
 * system device names to mount entries.
 */
#include "cmogstored.h"

/*
 * TODO: implement fine-grained locking since some network filesystems
 * can take a long time to stat()
 */
static pthread_mutex_t mnt_lock = PTHREAD_MUTEX_INITIALIZER;
/* global for the process */
static Hash_table *by_dev; /* maps (system) device IDs to a mount_entry */

struct mog_mntent {
	struct mount_entry *me;
	char util[8];
};

static void mount_entry_free(struct mount_entry *me)
{
	free(me->me_devname);
	free(me->me_mountdir);
	if (me->me_next)
		mount_entry_free(me->me_next);
	free(me);
}

static void me_free(void *entry)
{
	struct mog_mntent *mntent = entry;
	struct mount_entry *me = mntent->me;

	mount_entry_free(me);
	free(mntent);
}

static size_t me_dev_hash(const void *entry, size_t tablesize)
{
	const struct mog_mntent *mntent = entry;

	return mntent->me->me_dev % tablesize;
}

static bool me_dev_cmp(const void *a, const void *b)
{
	const struct mog_mntent *mntent_a = a;
	const struct mog_mntent *mntent_b = b;

	return mntent_a->me->me_dev == mntent_b->me->me_dev;
}

static void mnt_atexit(void)
{
	hash_free(by_dev);
}

static void mnt_once(void)
{
	by_dev = hash_initialize(7, NULL, me_dev_hash, me_dev_cmp, me_free);
	if (!by_dev)
		mog_oom();
	atexit(mnt_atexit);
}

static bool resolve_symlink(char **orig)
{
	char *p = canonicalize_filename_mode(*orig, CAN_EXISTING);

	if (p) {
		free(*orig);
		*orig = p;
		return true;
	}
	return false;
}

void mog_mnt_refresh(void)
{
	struct mount_entry *head = read_file_system_list(false);
	struct mount_entry *next;
	struct stat sb;
	union {
		const void *ptr;
		struct mog_mntent *oldent;
	} exist;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );

	if (by_dev)
		hash_clear(by_dev);
	else
		mnt_once();

	for ( ; head; head = next) {
		struct mog_mntent *mntent = NULL;

		next = head->me_next;
		head->me_next = NULL;
		if (head->me_type_malloced) {
			free(head->me_type);
			head->me_type_malloced = 0;
		}
		head->me_type = NULL;

		if (head->me_dummy)
			goto skip;

		mntent = xmalloc(sizeof(struct mog_mntent));
		mntent->me = head;
		mntent->util[0] = '-';
		mntent->util[1] = 0;

		/* the device number may not have been populated, do it */
		if (head->me_dev == (dev_t)-1) {
			if (stat(head->me_mountdir, &sb) != 0)
				goto skip;
			head->me_dev = sb.st_dev;
		}

		/*
		 * resolve symlinks for things that look like paths
		 * and skip dead symlinks
		 */
		if (head->me_devname[0] == '/') {
			if (lstat(head->me_devname, &sb) == 0 &&
			    S_ISLNK(sb.st_mode) &&
			    ! resolve_symlink(&head->me_devname))
				goto skip;
		}

		switch (hash_insert_if_absent(by_dev, mntent, &exist.ptr)) {
		case 0: {
			/* chain entries if they have multiple st_dev */
			struct mount_entry *me = exist.oldent->me;

			while (me->me_next)
				me = me->me_next;

			assert(me != head && "circular mount ref");
			me->me_next = head;
			free(mntent);
		}
			continue;
		case 1:
			continue;
		default: mog_oom();
		}
		assert(0 && "compiler bug?");
skip:
		free(mntent);
		mount_entry_free(head);
	}

	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
}

/*
 * Looks up a mount_entry by st_dev, returns NULL if nothing was found
 * Users may only acquire one mount entry at a time and MUST release it
 */
const struct mount_entry * mog_mnt_acquire(dev_t st_dev)
{
	struct mount_entry me = { .me_dev = st_dev };
	struct mog_mntent mntent = { .me = &me };
	struct mog_mntent *rv;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	rv = hash_lookup(by_dev, &mntent);

	/* user must release this via mog_mnt_release if non-NULL */
	if (rv) {
		struct mount_entry *rv_me = rv->me;

		/*
		 * if multiple entries match st_dev, favor the one
		 * with a leading slash
		 */
		while (rv_me && rv_me->me_devname[0] != '/')
			rv_me = rv_me->me_next;

		return rv_me ? rv_me : rv->me;
	}

	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
	return NULL;
}

/* releases the mount entry, allowing mog_mnt_acquire to be called again */
void mog_mnt_release(const struct mount_entry *me)
{
	struct mog_mntent mntent;
	const struct mog_mntent *check;
	struct mount_entry *check_me;
	union { const void *in; void *out; } deconst = { .in = me };

	mntent.me = deconst.out;
	check = hash_lookup(by_dev, &mntent);
	check_me = check->me;

	while (check_me->me_next && check_me != me)
		check_me = check_me->me_next;

	assert(check_me == me && "did not release acquired mount_entry");
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
}

struct mnt_update {
	const char *prefix;
	size_t prefixlen;
	const char *util;
};

/* attempt to resolve multiple mounts mapped to the same mount point */
static bool
update_util_chain(struct mog_mntent *mntent, struct mnt_update *update)
{
	struct mount_entry *me;
	size_t devnamelen;

	for (me = mntent->me->me_next; me; me = me->me_next) {
		devnamelen = strlen(me->me_devname);

		if (devnamelen < update->prefixlen)
			continue;
		if (memcmp(update->prefix, me->me_devname, update->prefixlen))
			continue;
		memcpy(mntent->util, update->util, 8);
	}

	return true;
}

static bool update_util_each(void *ent, void *upd)
{
	struct mog_mntent *mntent = ent;
	struct mnt_update *update = upd;
	const char *me_devname = mntent->me->me_devname;
	size_t devnamelen = strlen(me_devname);

	if (devnamelen < update->prefixlen)
		return update_util_chain(mntent, update);
	if (memcmp(update->prefix, me_devname, update->prefixlen) != 0)
		return update_util_chain(mntent, update);

	memcpy(mntent->util, update->util, 8);

	return true; /* continue */
}

/* this is O(mountpoints) for now */
void mog_mnt_update_util(const char *devsuffix, const char *util, size_t len)
{
	struct mnt_update update;

	assert(len >= 8 && "len too small");
	update.prefix = xasprintf("/dev/%s", devsuffix);
	update.prefixlen = strlen(update.prefix);
	update.util = util;

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	(void)hash_do_for_each(by_dev, update_util_each, &update);
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );
	mog_free(update.prefix);
}

char * mog_mnt_fetch_util(dev_t st_dev, char *dst, size_t len)
{
	struct mount_entry me = { .me_dev = st_dev };
	struct mog_mntent mntent = { .me = &me };
	struct mog_mntent *tmp;

	assert(len >= 8 && "len too small for dst");

	CHECK(int, 0, pthread_mutex_lock(&mnt_lock) );
	tmp = hash_lookup(by_dev, &mntent);
	if (tmp)
		memcpy(dst, tmp->util, 8);
	CHECK(int, 0, pthread_mutex_unlock(&mnt_lock) );

	return tmp ? dst : NULL;
}
